#ifndef _ASM_X86_HYGON_LMC_API_H
#define _ASM_X86_HYGON_LMC_API_H
#include <linux/bottom_half.h>

#ifdef CONFIG_X86_HYGON_LMC_SSE2_ON
void fpu_save_xmm0_3(void *to, const void *from, unsigned long len);
void fpu_restore_xmm0_3(void *to, const void *from, unsigned long len);

#define kernel_fpu_states_save fpu_save_xmm0_3
#define kernel_fpu_states_restore fpu_restore_xmm0_3

__must_check unsigned long copy_user_sse2_opt_string(void *to, const void *from,
						     unsigned long len);

#define MAX_FPU_CTX_SIZE 64
#define KERNEL_FPU_NONATOMIC_SIZE (2 * (MAX_FPU_CTX_SIZE))
#define copy_user_large_memory_generic_string copy_user_sse2_opt_string
#endif /* CONFIG_X86_HYGON_LMC_SSE2_ON */

#ifdef CONFIG_X86_HYGON_LMC_AVX2_ON
void fpu_save_ymm0_7(void *to, const void *from, unsigned long len);
void fpu_restore_ymm0_7(void *to, const void *from, unsigned long len);

#define kernel_fpu_states_save fpu_save_ymm0_7
#define kernel_fpu_states_restore fpu_restore_ymm0_7

__must_check unsigned long
copy_user_avx2_pf64_nt_string(void *to, const void *from, unsigned long len);

#define MAX_FPU_CTX_SIZE 256
#define KERNEL_FPU_NONATOMIC_SIZE (2 * (MAX_FPU_CTX_SIZE))
#define copy_user_large_memory_generic_string copy_user_avx2_pf64_nt_string
#endif /* CONFIG_X86_HYGON_LMC_AVX2_ON */

#ifdef CONFIG_USING_FPU_IN_KERNEL_NONATOMIC
int kernel_fpu_begin_nonatomic_mask(void);
void kernel_fpu_end_nonatomic(void);
void save_fpregs_to_fpkernelstate(struct fpu *kfpu);
unsigned long get_fpu_registers_pos(struct fpu *fpu, unsigned int off);

/* Code that is unaware of kernel_fpu_begin_nonatomic_mask() can use this */
static inline int kernel_fpu_begin_nonatomic(void)
{
	return kernel_fpu_begin_nonatomic_mask();
}

/*
 * It means we call kernel_fpu_end after kernel_fpu_begin_nonatomic
 * func, but before kernel_fpu_end_nonatomic
 */
static inline void check_using_kernel_fpu(void)
{
	if (boot_cpu_data.x86_vendor == X86_VENDOR_HYGON)
		WARN_ON_ONCE(test_thread_flag(TIF_USING_FPU_NONATOMIC));
}

/*
 * Kernel FPU state switching for scheduling.
 *
 * This is a two-stage process:
 *
 *  - switch_kernel_fpu_prepare() saves the old kernel fpu state.
 *    This is done within the context of the old process.
 *
 *  - switch_kernel_fpu_finish() restore new kernel fpu state.
 *
 * The kernel FPU context is only stored/restored for a user task in kernel
 * mode and PF_KTHREAD is used to distinguish between kernel and user threads.
 */
static inline void switch_kernel_fpu_prepare(struct task_struct *prev, int cpu)
{
	struct fpu *old_fpu = &prev->thread.fpu;

	if (boot_cpu_data.x86_vendor != X86_VENDOR_HYGON ||
	    !test_thread_flag(TIF_USING_FPU_NONATOMIC))
		return;

	if (static_cpu_has(X86_FEATURE_FPU) && !(prev->flags & PF_KTHREAD))
		save_fpregs_to_fpkernelstate(old_fpu);
}

/* Internal helper for switch_kernel_fpu_finish() and signal frame setup */
static inline void fpregs_restore_kernelregs(struct fpu *kfpu)
{
	kernel_fpu_states_restore(NULL,
				  (void *)get_fpu_registers_pos(kfpu,
								MAX_FPU_CTX_SIZE),
				  MAX_FPU_CTX_SIZE);
}

/* Loading of the complete FPU state immediately. */
static inline void switch_kernel_fpu_finish(struct task_struct *next)
{
	struct fpu *new_fpu = &next->thread.fpu;

	if ((next->flags & PF_KTHREAD) ||
	    boot_cpu_data.x86_vendor != X86_VENDOR_HYGON)
		return;

	if (cpu_feature_enabled(X86_FEATURE_FPU) &&
	    test_ti_thread_flag((struct thread_info *)next,
				TIF_USING_FPU_NONATOMIC))
		fpregs_restore_kernelregs(new_fpu);
}

unsigned int get_nt_block_copy_mini_len(void);
static inline bool Hygon_LMC_check(unsigned long len)
{
	unsigned int nt_blk_cpy_mini_len = get_nt_block_copy_mini_len();

	if (boot_cpu_data.x86_vendor == X86_VENDOR_HYGON &&
	    (nt_blk_cpy_mini_len && nt_blk_cpy_mini_len <= len &&
	     system_state == SYSTEM_RUNNING && !kernel_fpu_begin_nonatomic()))
		return true;
	else
		return false;
}

static inline unsigned long
copy_large_memory_generic_string(void *to, const void *from, unsigned long len)
{
	unsigned long ret;

	ret = copy_user_large_memory_generic_string(to, from, len);
	kernel_fpu_end_nonatomic();
	return ret;
}
#else
static inline void check_using_kernel_fpu(void)
{
}

static inline void switch_kernel_fpu_prepare(struct task_struct *prev, int cpu)
{
}

static inline void switch_kernel_fpu_finish(struct task_struct *next)
{
}

static inline bool Hygon_LMC_check(unsigned long len)
{
	return false;
}

static inline unsigned long
copy_large_memory_generic_string(void *to, const void *from, unsigned long len)
{
	return 0;
}
#endif /* CONFIG_USING_FPU_IN_KERNEL_NONATOMIC */

#endif /* _ASM_X86_HYGON_LMC_API_H */
